from __future__ import annotations

import datetime as dt
from collections.abc import Awaitable, Callable, Coroutine, Hashable, Iterable, Mapping
from enum import Enum
from logging import Logger
from pathlib import Path
from random import Random
from types import TracebackType
from typing import Any, ClassVar, Literal, Protocol, TypeVar, runtime_checkable
from zoneinfo import ZoneInfo

_T_contra = TypeVar("_T_contra", contravariant=True)


# basic
type Number = int | float
type Duration = Number | dt.timedelta
type StrMapping = Mapping[str, Any]
type TupleOrStrMapping = tuple[Any, ...] | StrMapping
type MaybeType[_T] = _T | type[_T]


# asyncio
type Coroutine1[_T] = Coroutine[Any, Any, _T]
type MaybeAwaitable[_T] = _T | Awaitable[_T]
type MaybeCoroutine1[_T] = _T | Coroutine1[_T]


# concurrent
type Parallelism = Literal["processes", "threads"]


# dataclasses
@runtime_checkable
class Dataclass(Protocol):
    """Protocol for `dataclass` classes."""

    __dataclass_fields__: ClassVar[dict[str, Any]]


# datetime
type DateOrDateTime = dt.date | dt.datetime


# enum
type EnumOrStr[_TEnum: Enum] = _TEnum | str


# iterables
type MaybeIterable[_T] = _T | Iterable[_T]
type IterableHashable[_THashable: Hashable] = (
    tuple[_THashable, ...] | frozenset[_THashable]
)
type MaybeIterableHashable[_THashable: Hashable] = (
    _THashable | IterableHashable[_THashable]
)


# logging
type LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
type LoggerOrName = Logger | str


# operator


class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, other: _T_contra, /) -> bool: ...  # pragma: no cover


class SupportsDunderGT(Protocol[_T_contra]):
    def __gt__(self, other: _T_contra, /) -> bool: ...  # pragma: no cover


SupportsRichComparison = SupportsDunderLT[Any] | SupportsDunderGT[Any]


# pathlib
type PathLike = Path | str
type PathLikeOrCallable = PathLike | Callable[[], PathLike]


# random
type Seed = int | float | str | bytes | bytearray | Random


# traceback
type ExcInfo = tuple[type[BaseException], BaseException, TracebackType]
type OptExcInfo = ExcInfo | tuple[None, None, None]


# zoneinfo
type ZoneInfoLike = ZoneInfo | str
type LocalOrZoneInfoLike = Literal["local"] | ZoneInfoLike


__all__ = [
    "Coroutine1",
    "Dataclass",
    "DateOrDateTime",
    "Duration",
    "EnumOrStr",
    "ExcInfo",
    "IterableHashable",
    "LocalOrZoneInfoLike",
    "LogLevel",
    "LoggerOrName",
    "MaybeAwaitable",
    "MaybeCoroutine1",
    "MaybeIterable",
    "MaybeIterableHashable",
    "MaybeType",
    "Number",
    "OptExcInfo",
    "Parallelism",
    "PathLike",
    "PathLikeOrCallable",
    "Seed",
    "StrMapping",
    "SupportsDunderGT",
    "SupportsDunderLT",
    "SupportsRichComparison",
    "TupleOrStrMapping",
    "ZoneInfoLike",
]



from __future__ import annotations

from collections.abc import Callable
from functools import cache as _cache
from functools import lru_cache as _lru_cache
from functools import partial as _partial
from typing import Any, TypeVar, cast, overload, override

_F = TypeVar("_F", bound=Callable[..., Any])
_T = TypeVar("_T")


def cache(func: _F, /) -> _F:
    """Typed version of `cache`."""
    typed_cache = cast(Callable[[_F], _F], _cache)
    return typed_cache(func)


##


@overload
def lru_cache(func: _F, /, *, max_size: int = ..., typed: bool = ...) -> _F: ...
@overload
def lru_cache(
    func: None = None, /, *, max_size: int = ..., typed: bool = ...
) -> Callable[[_F], _F]: ...
def lru_cache(
    func: _F | None = None, /, *, max_size: int = 128, typed: bool = False
) -> _F | Callable[[_F], _F]:
    """Typed version of `lru_cache`."""
    if func is None:
        result = partial(lru_cache, max_size=max_size, typed=typed)
        return cast(Callable[[_F], _F], result)
    wrapped = _lru_cache(maxsize=max_size, typed=typed)(func)
    return cast(Any, wrapped)


##


class partial(_partial[_T]):  # noqa: N801
    """Partial which accepts Ellipsis for positional arguments."""

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> _T:
        iter_args = iter(args)
        head = (next(iter_args) if arg is ... else arg for arg in self.args)
        return self.func(*head, *iter_args, **{**self.keywords, **kwargs})


__all__ = ["cache", "lru_cache", "partial"]



from __future__ import annotations

from asyncio import sleep

from utilities.atools import call_memoized


class TestCallMemoized:
    async def test_main(self) -> None:
        i = 0

        async def increment() -> int:
            nonlocal i
            i += 1
            return i

        assert (await call_memoized(increment)) == 1
        assert (await call_memoized(increment)) == 2
        for _ in range(2):
            assert (await call_memoized(increment, 0.01)) == 3
        await sleep(0.01)
        assert (await call_memoized(increment)) == 4
        assert (await call_memoized(increment)) == 5
        for _ in range(2):
            assert (await call_memoized(increment, 0.01)) == 6



from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import (
    CancelledError,
    Lock,
    PriorityQueue,
    Queue,
    QueueEmpty,
    Semaphore,
    StreamReader,
    Task,
    TaskGroup,
    create_subprocess_shell,
    create_task,
    sleep,
    timeout,
)
from collections.abc import AsyncIterator, Hashable
from contextlib import asynccontextmanager, suppress
from dataclasses import dataclass, field
from io import StringIO
from subprocess import PIPE
from sys import stderr, stdout
from typing import TYPE_CHECKING, Any, Generic, Self, TextIO, TypeVar, cast, override

from utilities.datetime import datetime_duration_to_float
from utilities.functions import ensure_int, ensure_not_none
from utilities.types import SupportsRichComparison

if TYPE_CHECKING:
    from asyncio import _CoroutineLike
    from asyncio.subprocess import Process
    from collections.abc import Sequence
    from contextvars import Context
    from types import TracebackType

    from utilities.types import Duration

_T = TypeVar("_T")
_THashable = TypeVar("_THashable", bound=Hashable)
_TSupportsRichComparison = TypeVar(
    "_TSupportsRichComparison", bound=SupportsRichComparison
)


class BoundedTaskGroup(TaskGroup):
    """Task group with an internal limiter."""

    _semaphore: Semaphore | None

    @override
    def __init__(self, *, max_tasks: int | None = None) -> None:
        super().__init__()
        self._semaphore = None if max_tasks is None else Semaphore(max_tasks)

    @override
    def create_task(
        self,
        coro: _CoroutineLike[_T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> Task[_T]:
        if self._semaphore is None:
            return super().create_task(coro, name=name, context=context)

        async def wrapped(semaphore: Semaphore, coro: _CoroutineLike[_T], /) -> _T:
            async with semaphore:
                return await cast(Any, coro)

        return super().create_task(
            wrapped(self._semaphore, coro), name=name, context=context
        )


##


@dataclass(kw_only=True)
class QueueProcessor(ABC, Generic[_T]):
    """Process a set of items in a queue."""

    queue_type: type[Queue[_T]] = field(default=Queue, repr=False)
    queue_max_size: int | None = field(default=None, repr=False)
    _lock: Lock = field(init=False, repr=False)
    _queue: Queue[_T] = field(init=False, repr=False)
    _task: Task[None] | None = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self._lock = Lock()
        self._queue = self.queue_type(
            maxsize=0 if self.queue_max_size is None else self.queue_max_size
        )
        self._task = None

    async def __aenter__(self) -> Self:
        """Start the server."""
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """Stop the server."""
        _ = (exc_type, exc_value, traceback)
        await self.stop()

    def __del__(self) -> None:
        try:
            task = self._task
        except AttributeError:
            return
        if task is None:
            return
        with suppress(RuntimeError):
            _ = task.cancel()

    def __len__(self) -> int:
        return self._queue.qsize()

    def empty(self) -> bool:
        """Check if the queue is empty."""
        return self._queue.empty()

    def enqueue(self, *items: _T) -> None:
        """Enqueue a set items."""
        for item in items:
            self._queue.put_nowait(item)

    @classmethod
    async def new(cls, *args: _T, **kwargs: Any) -> Self:
        """Create and start ."""
        self = cls(**kwargs)
        self.enqueue(*args)
        await self.start()
        return self

    async def run_until_empty(self) -> None:
        """Run the processor until the queue is empty."""
        while not self.empty():
            _ = await self._get_and_run()

    async def start(self) -> None:
        """Start the processor."""
        if self._task is None:
            self._task = create_task(self._loop())

    async def stop(self) -> None:
        """Stop the processor."""
        if self._task is None:
            return
        _ = self._task.cancel()
        await self.run_until_empty()
        with suppress(CancelledError):
            await self._task
        self._task = None

    async def _get_and_run(self) -> None:
        """Get the next item and run the processor."""
        (item,) = await self._get_items(max_size=1)
        await self._run(item)

    async def _get_items(self, *, max_size: int | None = None) -> Sequence[_T]:
        """Get items from the queue; if empty then wait."""
        try:
            return await get_items(self._queue, max_size=max_size, lock=self._lock)
        except RuntimeError as error:  # pragma: no cover
            if error.args[0] == "Event loop is closed":
                return []
            raise

    async def _get_items_nowait(self, *, max_size: int | None = None) -> Sequence[_T]:
        """Get items from the queue; no waiting."""
        return await get_items_nowait(self._queue, max_size=max_size, lock=self._lock)

    async def _loop(self, /) -> None:
        """Loop the processor."""
        while True:
            try:
                await self._get_and_run()
            except RuntimeError as error:  # pragma: no cover
                if error.args[0] == "Event loop is closed":
                    return
                raise

    @abstractmethod
    async def _run(self, item: _T, /) -> None:
        """Run the processor on the first item."""
        raise NotImplementedError(item)  # pragma: no cover


##


class UniquePriorityQueue(PriorityQueue[tuple[_TSupportsRichComparison, _THashable]]):
    """Priority queue with unique tasks."""

    @override
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._set: set[_THashable] = set()

    @override
    def _get(self) -> tuple[_TSupportsRichComparison, _THashable]:
        item = super()._get()
        _, value = item
        self._set.remove(value)
        return item

    @override
    def _put(self, item: tuple[_TSupportsRichComparison, _THashable]) -> None:
        _, value = item
        if value not in self._set:
            super()._put(item)
            self._set.add(value)


class UniqueQueue(Queue[_THashable]):
    """Queue with unique tasks."""

    @override
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._set: set[_THashable] = set()

    @override
    def _get(self) -> _THashable:
        item = super()._get()
        self._set.remove(item)
        return item

    @override
    def _put(self, item: _THashable) -> None:
        if item not in self._set:
            super()._put(item)
            self._set.add(item)


##


async def get_items(
    queue: Queue[_T], /, *, max_size: int | None = None, lock: Lock | None = None
) -> list[_T]:
    """Get items from a queue; if empty then wait."""
    items = [await queue.get()]
    max_size_use = None if max_size is None else (max_size - 1)
    if lock is None:
        items.extend(await get_items_nowait(queue, max_size=max_size_use))
    else:
        async with lock:
            items.extend(await get_items_nowait(queue, max_size=max_size_use))
    return items


async def get_items_nowait(
    queue: Queue[_T], /, *, max_size: int | None = None, lock: Lock | None = None
) -> list[_T]:
    """Get items from a queue; no waiting."""
    if lock is None:
        return _get_items_nowait_core(queue, max_size=max_size)
    async with lock:
        return _get_items_nowait_core(queue, max_size=max_size)


def _get_items_nowait_core(
    queue: Queue[_T], /, *, max_size: int | None = None
) -> list[_T]:
    """Get all the items from a queue; no waiting."""
    items: list[_T] = []
    if max_size is None:
        while True:
            try:
                items.append(queue.get_nowait())
            except QueueEmpty:
                break
    else:
        while len(items) < max_size:
            try:
                items.append(queue.get_nowait())
            except QueueEmpty:
                break
    return items


##


async def sleep_dur(*, duration: Duration | None = None) -> None:
    """Sleep which accepts durations."""
    if duration is None:
        return
    await sleep(datetime_duration_to_float(duration))


##


@dataclass(kw_only=True, slots=True)
class StreamCommandOutput:
    process: Process
    stdout: str
    stderr: str

    @property
    def return_code(self) -> int:
        return ensure_int(self.process.returncode)  # skipif-not-windows


async def stream_command(cmd: str, /) -> StreamCommandOutput:
    """Run a shell command asynchronously and stream its output in real time."""
    process = await create_subprocess_shell(  # skipif-not-windows
        cmd, stdout=PIPE, stderr=PIPE
    )
    proc_stdout = ensure_not_none(  # skipif-not-windows
        process.stdout, desc="process.stdout"
    )
    proc_stderr = ensure_not_none(  # skipif-not-windows
        process.stderr, desc="process.stderr"
    )
    ret_stdout = StringIO()  # skipif-not-windows
    ret_stderr = StringIO()  # skipif-not-windows
    async with TaskGroup() as tg:  # skipif-not-windows
        _ = tg.create_task(_stream_one(proc_stdout, stdout, ret_stdout))
        _ = tg.create_task(_stream_one(proc_stderr, stderr, ret_stderr))
    _ = await process.wait()  # skipif-not-windows
    return StreamCommandOutput(  # skipif-not-windows
        process=process, stdout=ret_stdout.getvalue(), stderr=ret_stderr.getvalue()
    )


async def _stream_one(
    input_: StreamReader, out_stream: TextIO, ret_stream: StringIO, /
) -> None:
    """Asynchronously read from a stream and write to the target output stream."""
    while True:  # skipif-not-windows
        line = await input_.readline()
        if not line:
            break
        decoded = line.decode()
        _ = out_stream.write(decoded)
        out_stream.flush()
        _ = ret_stream.write(decoded)


##


@asynccontextmanager
async def timeout_dur(
    *, duration: Duration | None = None, error: type[Exception] = TimeoutError
) -> AsyncIterator[None]:
    """Timeout context manager which accepts durations."""
    delay = None if duration is None else datetime_duration_to_float(duration)
    try:
        async with timeout(delay):
            yield
    except TimeoutError:
        raise error from None


__all__ = [
    "BoundedTaskGroup",
    "QueueProcessor",
    "StreamCommandOutput",
    "UniquePriorityQueue",
    "UniqueQueue",
    "get_items",
    "get_items_nowait",
    "sleep_dur",
    "stream_command",
    "timeout_dur",
]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Mapping, Sequence
from collections.abc import Set as AbstractSet
from dataclasses import asdict, dataclass
from typing import Any, TypeVar, cast, override

import utilities.math
from utilities.datetime import (
    AreEqualDatesOrDateTimesError,
    AreEqualDateTimesError,
    are_equal_dates_or_datetimes,
)
from utilities.functions import is_dataclass_instance
from utilities.iterables import SortIterableError, sort_iterable
from utilities.reprlib import get_repr
from utilities.types import Dataclass, DateOrDateTime, Number

_T = TypeVar("_T")


def is_equal(
    x: Any,
    y: Any,
    /,
    *,
    rel_tol: float | None = None,
    abs_tol: float | None = None,
    extra: Mapping[type[_T], Callable[[_T, _T], bool]] | None = None,
) -> bool:
    """Check if two objects are equal."""
    if type(x) is type(y):
        # extra
        if extra is not None:
            try:
                cmp = next(cmp for cls, cmp in extra.items() if isinstance(x, cls))
            except StopIteration:
                pass
            else:
                x = cast(_T, x)
                y = cast(_T, y)
                return cmp(x, y)

        # singletons
        if isinstance(x, int | float):
            y = cast(Number, y)
            return utilities.math.is_equal(x, y, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, str):  # else Sequence
            y = cast(str, y)
            return x == y
        if isinstance(x, dt.date | dt.datetime):
            y = cast(DateOrDateTime, y)
            try:
                return are_equal_dates_or_datetimes(x, y)
            except (AreEqualDateTimesError, AreEqualDatesOrDateTimesError):
                return False
        if is_dataclass_instance(x):
            y = cast(Dataclass, y)
            x_values = asdict(x)
            y_values = asdict(y)
            return is_equal(x_values, y_values, rel_tol=rel_tol, abs_tol=abs_tol)

        # collections
        if isinstance(x, Mapping):
            y = cast(Mapping[Any, Any], y)
            x_keys = set(x)
            y_keys = set(y)
            if not is_equal(x_keys, y_keys, rel_tol=rel_tol, abs_tol=abs_tol):
                return False
            x_values = [x[i] for i in x]
            y_values = [y[i] for i in x]
            return is_equal(x_values, y_values, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, AbstractSet):
            y = cast(AbstractSet[Any], y)
            try:
                x_sorted = sort_iterable(x)
                y_sorted = sort_iterable(y)
            except SortIterableError as error:
                raise IsEqualError(x=error.x, y=error.y) from None
            return is_equal(x_sorted, y_sorted, rel_tol=rel_tol, abs_tol=abs_tol)
        if isinstance(x, Sequence):
            y = cast(Sequence[Any], y)
            if len(x) != len(y):
                return False
            return all(
                is_equal(x_i, y_i, rel_tol=rel_tol, abs_tol=abs_tol)
                for x_i, y_i in zip(x, y, strict=True)
            )

    if isinstance(x, int | float) and isinstance(y, int | float):
        return utilities.math.is_equal(x, y, rel_tol=rel_tol, abs_tol=abs_tol)

    return (type(x) is type(y)) and (x == y)


@dataclass(kw_only=True, slots=True)
class IsEqualError(Exception):
    x: Any
    y: Any

    @override
    def __str__(self) -> str:
        return f"Unable to sort {get_repr(self.x)} and {get_repr(self.y)}"  # pragma: no cover


__all__ = ["IsEqualError", "is_equal"]



from __future__ import annotations

import datetime as dt
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from types import NoneType
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, Self
from uuid import UUID

from hypothesis import given
from hypothesis.strategies import DataObject, data, just, none, sampled_from
from pytest import mark, param

from tests.test_typing_funcs.no_future import (
    DataClassNestedNoFutureInnerThenOuterInner,
    DataClassNestedNoFutureInnerThenOuterOuter,
    DataClassNestedNoFutureOuterThenInnerInner,
    DataClassNestedNoFutureOuterThenInnerOuter,
)
from tests.test_typing_funcs.with_future import (
    DataClassNestedWithFutureInnerThenOuterInner,
    DataClassNestedWithFutureInnerThenOuterOuter,
    DataClassNestedWithFutureOuterThenInnerInner,
    DataClassNestedWithFutureOuterThenInnerOuter,
    DataClassWithDate,
    DataClassWithInt,
    DataClassWithIntNullable,
    DataClassWithListInts,
    DataClassWithLiteral,
    DataClassWithNone,
    DataClassWithPath,
    DataClassWithSentinel,
    DataClassWithStr,
    DataClassWithTimeDelta,
    DataClassWithUUID,
)
from utilities.sentinel import Sentinel
from utilities.types import LogLevel, Parallelism
from utilities.typing import (
    contains_self,
    get_args,
    get_type_hints,
    is_dict_type,
    is_frozenset_type,
    is_list_type,
    is_literal_type,
    is_mapping_type,
    is_namedtuple_class,
    is_namedtuple_instance,
    is_optional_type,
    is_sequence_type,
    is_set_type,
    is_union_type,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class TestContainsSelf:
    @mark.parametrize("obj", [param(Self), param(Self | None)])
    def test_main(self, *, obj: Any) -> None:
        assert contains_self(obj)


class TestGetArgs:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(dict[int, int], (int, int)),
            param(frozenset[int], (int,)),
            param(int | None, (int,)),
            param(int | str, (int, str)),
            param(list[int], (int,)),
            param(Literal["a", "b", "c"], ("a", "b", "c")),
            param(Mapping[int, int], (int, int)),
            param(Sequence[int], (int,)),
            param(set[int], (int,)),
            param(LogLevel, ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")),
            param(Parallelism, ("processes", "threads")),
        ],
    )
    def test_main(self, *, obj: Any, expected: tuple[Any, ...]) -> None:
        result = get_args(obj)
        assert result == expected


class TestGetTypeHints:
    @given(data=data())
    def test_date(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            date: dt.date

        cls = data.draw(sampled_from([Example, DataClassWithDate]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"date": dt.date}
        assert hints == expected

    @given(data=data())
    def test_int(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            int_: int

        cls = data.draw(sampled_from([Example, DataClassWithInt]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"int_": int}
        assert hints == expected

    @given(data=data())
    def test_int_nullable(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            int_: int | None = None

        cls = data.draw(sampled_from([Example, DataClassWithIntNullable]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"int_": int | None}
        assert hints == expected

    @given(data=data())
    def test_list(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            ints: list[int]

        cls = data.draw(sampled_from([Example, DataClassWithListInts]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"ints": list[int]}
        assert hints == expected

    def test_nested_local(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner

        @dataclass(kw_only=True, slots=True)
        class Inner:
            int_: int

        hints = get_type_hints(Outer, localns=locals())
        expected = {"inner": Inner}
        assert hints == expected

    def test_nested_no_future_inner_then_outer(self) -> None:
        hints = get_type_hints(
            DataClassNestedNoFutureInnerThenOuterOuter, globalns=globals()
        )
        expected = {"inner": DataClassNestedNoFutureInnerThenOuterInner}
        assert hints == expected

    def test_nested_no_future_outer_then_inner(self) -> None:
        hints = get_type_hints(
            DataClassNestedNoFutureOuterThenInnerOuter, globalns=globals()
        )
        expected = {"inner": DataClassNestedNoFutureOuterThenInnerInner}
        assert hints == expected

    def test_nested_with_future_inner_then_outer(self) -> None:
        hints = get_type_hints(
            DataClassNestedWithFutureInnerThenOuterOuter, globalns=globals()
        )
        expected = {"inner": DataClassNestedWithFutureInnerThenOuterInner}
        assert hints == expected

    def test_nested_with_future_outer_then_inner(self) -> None:
        hints = get_type_hints(
            DataClassNestedWithFutureOuterThenInnerOuter, globalns=globals()
        )
        expected = {"inner": DataClassNestedWithFutureOuterThenInnerInner}
        assert hints == expected

    @given(data=data())
    def test_literal(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: Literal["true", "false"]

        cls = data.draw(sampled_from([Example, DataClassWithLiteral]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"truth": Literal["true", "false"]}
        assert hints == expected

    @given(data=data())
    def test_none(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            none: None

        cls = data.draw(sampled_from([Example, DataClassWithNone]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"none": NoneType}
        assert hints == expected

    @given(data=data())
    def test_path(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            path: Path

        cls = data.draw(sampled_from([Example, DataClassWithPath]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"path": Path}
        assert hints == expected

    @given(data=data())
    def test_sentinel(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            sentinel: Sentinel

        cls = data.draw(sampled_from([Example, DataClassWithSentinel]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"sentinel": Sentinel}
        assert hints == expected

    @given(data=data())
    def test_str(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            str_: str

        cls = data.draw(sampled_from([Example, DataClassWithStr]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"str_": str}
        assert hints == expected

    @given(data=data())
    def test_timedelta(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            timedelta: dt.timedelta

        cls = data.draw(sampled_from([Example, DataClassWithTimeDelta]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"timedelta": dt.timedelta}
        assert hints == expected

    @given(data=data())
    def test_uuid(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            uuid: UUID

        cls = data.draw(sampled_from([Example, DataClassWithUUID]))
        globalns = data.draw(just(globals()) | none())
        localns = data.draw(just(locals()) | none())
        hints = get_type_hints(cls, globalns=globalns, localns=localns)
        expected = {"uuid": UUID}
        assert hints == expected

    def test_unresolved(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner

        @dataclass(kw_only=True, slots=True)
        class Inner:
            int_: int

        hints = get_type_hints(Outer)
        expected = {"inner": "Inner"}
        assert hints == expected


class TestIsAnnotationOfType:
    @mark.parametrize(
        ("func", "obj", "expected"),
        [
            param(is_dict_type, dict[int, int], True),
            param(is_dict_type, list[int], False),
            param(is_frozenset_type, frozenset[int], True),
            param(is_frozenset_type, list[int], False),
            param(is_list_type, list[int], True),
            param(is_list_type, set[int], False),
            param(is_mapping_type, Mapping[int, int], True),
            param(is_mapping_type, list[int], False),
            param(is_literal_type, Literal["a", "b", "c"], True),
            param(is_literal_type, list[int], False),
            param(is_optional_type, int | None, True),
            param(is_optional_type, int | str, False),
            param(is_optional_type, list[int], False),
            param(is_optional_type, list[int] | None, True),
            param(is_optional_type, Literal["a", "b", "c"], False),
            param(is_optional_type, Literal["a", "b", "c"] | None, True),
            param(is_sequence_type, Sequence[int], True),
            param(is_sequence_type, list[int], False),
            param(is_set_type, list[int], False),
            param(is_union_type, int | str, True),
            param(is_union_type, list[int], False),
        ],
    )
    def test_main(
        self, *, func: Callable[[Any], bool], obj: Any, expected: bool
    ) -> None:
        assert func(obj) is expected


class TestIsNamedTuple:
    def test_main(self) -> None:
        class Example(NamedTuple):
            x: int

        assert is_namedtuple_class(Example)
        assert is_namedtuple_instance(Example(x=0))

    def test_class(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        assert not is_namedtuple_class(Example)
        assert not is_namedtuple_instance(Example(x=0))



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from pathlib import Path
from typing import TypeVar

from hypothesis import given
from hypothesis.strategies import sampled_from
from pytest import mark, param

from utilities.datetime import ZERO_TIME
from utilities.types import Dataclass, Duration, Number, PathLike


class TestDataClassProtocol:
    def test_main(self) -> None:
        TDataclass = TypeVar("TDataclass", bound=Dataclass)

        def identity(x: TDataclass, /) -> TDataclass:
            return x

        @dataclass(kw_only=True, slots=True)
        class Example:
            x: None = None

        _ = identity(Example())


class TestDuration:
    @given(x=sampled_from([0, 0.0, ZERO_TIME]))
    def test_success(self, *, x: Duration) -> None:
        assert isinstance(x, int | float | dt.timedelta)

    def test_error(self) -> None:
        assert not isinstance("0", int | float | dt.timedelta)


class TestNumber:
    @given(x=sampled_from([0, 0.0]))
    def test_ok(self, *, x: Number) -> None:
        assert isinstance(x, int | float)

    def test_error(self) -> None:
        assert not isinstance(None, int | float)


class TestPathLike:
    @mark.parametrize("path", [param(Path.home()), param("~")])
    def test_main(self, *, path: PathLike) -> None:
        assert isinstance(path, Path | str)

    def test_error(self) -> None:
        assert not isinstance(None, Path | str)



from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import integers, lists, sampled_from, sets
from pytest import mark, param

from utilities.hypothesis import text_ascii
from utilities.sentinel import sentinel
from utilities.text import (
    join_strs,
    repr_encode,
    snake_case,
    split_str,
    str_encode,
    strip_and_dedent,
)


class TestReprEncode:
    @given(n=integers())
    def test_main(self, *, n: int) -> None:
        result = repr_encode(n)
        expected = repr(n).encode()
        assert result == expected


class TestSplitStrAndJoinStr:
    @mark.parametrize(
        ("text", "texts"),
        [
            param("", [""]),
            param("1", ["1"]),
            param("1,2", ["1", "2"]),
            param(",", ["", ""]),
            param(str(sentinel), []),
        ],
    )
    def test_main(self, *, text: str, texts: list[str]) -> None:
        assert split_str(text) == texts
        assert join_strs(texts) == text

    @given(texts=lists(text_ascii()))
    def test_generic(self, *, texts: list[str]) -> None:
        assert split_str(join_strs(texts)) == texts

    @given(texts=sets(text_ascii()))
    def test_sort(self, *, texts: set[str]) -> None:
        assert split_str(join_strs(texts, sort=True)) == sorted(texts)


class TestSnakeCase:
    @given(
        case=sampled_from([
            ("API", "api"),
            ("APIResponse", "api_response"),
            ("ApplicationController", "application_controller"),
            ("Area51Controller", "area51_controller"),
            ("FreeBSD", "free_bsd"),
            ("HTML", "html"),
            ("HTMLTidy", "html_tidy"),
            ("HTMLTidyGenerator", "html_tidy_generator"),
            ("HTMLVersion", "html_version"),
            ("NoHTML", "no_html"),
            ("One   Two", "one_two"),
            ("One  Two", "one_two"),
            ("One Two", "one_two"),
            ("OneTwo", "one_two"),
            ("One_Two", "one_two"),
            ("One__Two", "one_two"),
            ("One___Two", "one_two"),
            ("Product", "product"),
            ("SpecialGuest", "special_guest"),
            ("Text", "text"),
            ("Text123", "text123"),
            ("_APIResponse_", "_api_response_"),
            ("_API_", "_api_"),
            ("__APIResponse__", "_api_response_"),
            ("__API__", "_api_"),
            ("__impliedVolatility_", "_implied_volatility_"),
            ("_itemID", "_item_id"),
            ("_lastPrice__", "_last_price_"),
            ("_symbol", "_symbol"),
            ("aB", "a_b"),
            ("changePct", "change_pct"),
            ("changePct_", "change_pct_"),
            ("impliedVolatility", "implied_volatility"),
            ("lastPrice", "last_price"),
            ("memMB", "mem_mb"),
            ("sizeX", "size_x"),
            ("symbol", "symbol"),
            ("testNTest", "test_n_test"),
            ("text", "text"),
            ("text123", "text123"),
        ])
    )
    def test_main(self, *, case: tuple[str, str]) -> None:
        text, expected = case
        result = snake_case(text)
        assert result == expected


class TestStrEncode:
    @given(n=integers())
    def test_main(self, *, n: int) -> None:
        result = str_encode(n)
        expected = str(n).encode()
        assert result == expected


class TestStripAndDedent:
    @mark.parametrize("trailing", [param(True), param(False)])
    def test_main(self, *, trailing: bool) -> None:
        text = """
               This is line 1.
               This is line 2.
               """
        result = strip_and_dedent(text, trailing=trailing)
        expected = "This is line 1.\nThis is line 2." + ("\n" if trailing else "")
        assert result == expected



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, cast, overload, override

import luigi
from luigi import Parameter, PathParameter, Target, Task
from luigi import build as _build

from utilities.datetime import EPOCH_UTC

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import Iterable

    from luigi.execution_summary import LuigiRunResult

    from utilities.logging import LogLevel
    from utilities.types import PathLike


# parameters


class DateHourParameter(luigi.DateHourParameter):
    """A parameter which takes the value of an hourly `dt.datetime`."""

    def __init__(self, interval: int = 1, **kwargs: Any) -> None:
        super().__init__(interval, EPOCH_UTC, **kwargs)

    @override
    def normalize(self, dt: dt.datetime | str) -> dt.datetime:
        from utilities.whenever import ensure_zoned_datetime

        return ensure_zoned_datetime(dt)

    @override
    def parse(self, s: str) -> dt.datetime:
        from utilities.whenever import parse_zoned_datetime

        return parse_zoned_datetime(s)

    @override
    def serialize(self, dt: dt.datetime) -> str:
        from utilities.whenever import serialize_zoned_datetime

        return serialize_zoned_datetime(dt)


class DateMinuteParameter(luigi.DateMinuteParameter):
    """A parameter which takes the value of a minutely `dt.datetime`."""

    def __init__(self, interval: int = 1, **kwargs: Any) -> None:
        super().__init__(interval=interval, start=EPOCH_UTC, **kwargs)

    @override
    def normalize(self, dt: dt.datetime | str) -> dt.datetime:
        from utilities.whenever import ensure_zoned_datetime

        return ensure_zoned_datetime(dt)

    @override
    def parse(self, s: str) -> dt.datetime:
        from utilities.whenever import parse_zoned_datetime

        return parse_zoned_datetime(s)

    @override
    def serialize(self, dt: dt.datetime) -> str:
        from utilities.whenever import serialize_zoned_datetime

        return serialize_zoned_datetime(dt)


class DateSecondParameter(luigi.DateSecondParameter):
    """A parameter which takes the value of a secondly `dt.datetime`."""

    def __init__(self, interval: int = 1, **kwargs: Any) -> None:
        super().__init__(interval, EPOCH_UTC, **kwargs)

    @override
    def normalize(self, dt: dt.datetime | str) -> dt.datetime:
        from utilities.whenever import ensure_zoned_datetime

        return ensure_zoned_datetime(dt)

    @override
    def parse(self, s: str) -> dt.datetime:
        from utilities.whenever import parse_zoned_datetime

        return parse_zoned_datetime(s)

    @override
    def serialize(self, dt: dt.datetime) -> str:
        from utilities.whenever import serialize_zoned_datetime

        return serialize_zoned_datetime(dt)


class TimeParameter(Parameter):
    """A parameter which takes the value of a `dt.time`."""

    @override
    def normalize(self, x: dt.time | str) -> dt.time:
        from utilities.whenever import ensure_time

        return ensure_time(x)

    @override
    def parse(self, x: str) -> dt.time:
        from utilities.whenever import parse_time

        return parse_time(x)

    @override
    def serialize(self, x: dt.time) -> str:
        from utilities.whenever import serialize_time

        return serialize_time(x)


# targets


class PathTarget(Target):
    """A local target whose `path` attribute is a Pathlib instance."""

    def __init__(self, path: PathLike, /) -> None:
        super().__init__()
        self.path = Path(path)

    @override
    def exists(self) -> bool:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Check if the target exists."""
        return self.path.exists()


# tasks


class ExternalTask(ABC, luigi.ExternalTask):
    """An external task with `exists()` defined here."""

    @abstractmethod
    def exists(self) -> bool:
        """Predicate on which the external task is deemed to exist."""
        msg = f"{self=}"  # pragma: no cover
        raise NotImplementedError(msg)  # pragma: no cover

    @override
    def output(self) -> _ExternalTaskDummyTarget:  # pyright: ignore[reportIncompatibleMethodOverride]
        return _ExternalTaskDummyTarget(self)


class _ExternalTaskDummyTarget(Target):
    """Dummy target for `ExternalTask`."""

    def __init__(self, task: ExternalTask, /) -> None:
        super().__init__()
        self._task = task

    @override
    def exists(self) -> bool:  # pyright: ignore[reportIncompatibleMethodOverride]
        return self._task.exists()


class ExternalFile(ExternalTask):
    """Await an external file on the local disk."""

    path: Path = cast(Any, PathParameter())

    @override
    def exists(self) -> bool:
        return self.path.exists()


# functions


@overload
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: Literal[False] = False,
    local_scheduler: bool = False,
    log_level: LogLevel | None = None,
    workers: int | None = None,
) -> bool: ...
@overload
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: Literal[True],
    local_scheduler: bool = False,
    log_level: LogLevel | None = None,
    workers: int | None = None,
) -> LuigiRunResult: ...
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: bool = False,
    local_scheduler: bool = False,
    log_level: LogLevel | None = None,
    workers: int | None = None,
) -> bool | LuigiRunResult:
    """Build a set of tasks."""
    return _build(
        task,
        detailed_summary=detailed_summary,
        local_scheduler=local_scheduler,
        **({} if log_level is None else {"log_level": log_level}),
        **({} if workers is None else {"workers": workers}),
    )


__all__ = [
    "DateHourParameter",
    "DateMinuteParameter",
    "DateSecondParameter",
    "ExternalFile",
    "ExternalTask",
    "PathTarget",
    "TimeParameter",
    "build",
]



from __future__ import annotations

from dataclasses import dataclass
from re import search
from typing import TYPE_CHECKING

from hypothesis import HealthCheck, assume, given, settings
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    dates,
    datetimes,
    just,
    permutations,
    sampled_from,
    sets,
    timedeltas,
    timezones,
    tuples,
)
from pytest import raises

from utilities.datetime import ZERO_TIME
from utilities.hypothesis import assume_does_not_raise, pairs, zoned_datetimes
from utilities.period import (
    Period,
    _DateOrDateTime,
    _PeriodAsTimeZoneInapplicableError,
    _PeriodDateAndDateTimeMixedError,
    _PeriodDateContainsDateTimeError,
    _PeriodDateTimeContainsDateError,
    _PeriodInvalidError,
    _PeriodMaxDurationError,
    _PeriodMinDurationError,
    _PeriodNaiveDateTimeError,
    _PeriodReqDurationError,
    _PeriodTimeZoneInapplicableError,
    _PeriodTimeZoneNonUniqueError,
    _TPeriod,
)

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import Callable
    from zoneinfo import ZoneInfo


class TestPeriod:
    @given(dates=pairs(dates(), sorted=True), duration=timedeltas())
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    def test_add(
        self, *, dates: tuple[dt.date, dt.date], duration: dt.timedelta
    ) -> None:
        start, end = dates
        with assume_does_not_raise(OverflowError):
            adj_start, adj_end = start + duration, end + duration
        period = Period(start, end)
        result = period + duration
        expected = Period(adj_start, adj_end)
        assert result == expected

    @given(data=data(), time_zone1=timezones(), time_zone2=timezones())
    def test_astimezone(
        self, *, data: DataObject, time_zone1: ZoneInfo, time_zone2: ZoneInfo
    ) -> None:
        start, end = data.draw(
            pairs(zoned_datetimes(time_zone=time_zone1), sorted=True)
        )
        with assume_does_not_raise(OverflowError):
            adj_start, adj_end = (
                start.astimezone(time_zone2),
                end.astimezone(time_zone2),
            )
        period = Period(start, end)
        result = period.astimezone(time_zone2)
        expected = Period(adj_start, adj_end)
        assert result == expected

    @given(date=dates(), dates=pairs(dates(), sorted=True))
    def test_contain_date(
        self, *, date: dt.date, dates: tuple[dt.date, dt.date]
    ) -> None:
        start, end = dates
        period = Period(start, end)
        result = date in period
        expected = start <= date <= end
        assert result is expected

    @given(
        datetime=zoned_datetimes(time_zone=timezones()),
        datetimes=pairs(zoned_datetimes(time_zone=timezones()), sorted=True),
    )
    def test_contain_datetime(
        self, *, datetime: dt.datetime, datetimes: tuple[dt.datetime, dt.datetime]
    ) -> None:
        start, end = datetimes
        period = Period(start, end)
        result = datetime in period
        expected = start <= datetime <= end
        assert result is expected

    @given(dates=pairs(dates(), sorted=True))
    def test_dates(self, *, dates: tuple[dt.date, dt.date]) -> None:
        start, end = dates
        _ = Period(start, end)

    @given(datetimes=pairs(zoned_datetimes(time_zone=timezones()), sorted=True))
    def test_datetimes(self, *, datetimes: tuple[dt.datetime, dt.datetime]) -> None:
        start, end = datetimes
        _ = Period(start, end)

    @given(dates=pairs(dates(), sorted=True))
    def test_duration(self, *, dates: tuple[dt.date, dt.date]) -> None:
        start, end = dates
        period = Period(start, end)
        assert period.duration == (end - start)

    @given(dates=pairs(dates(), sorted=True))
    def test_hashable(self, *, dates: tuple[dt.date, dt.date]) -> None:
        start, end = dates
        period = Period(start, end)
        _ = hash(period)

    @given(
        case=tuples(dates(), just("date")) | tuples(zoned_datetimes(), just("datetime"))
    )
    def test_kind(self, *, case: tuple[dt.date, _DateOrDateTime]) -> None:
        date_or_datetime, kind = case
        period = Period(date_or_datetime, date_or_datetime)
        assert period.kind == kind

    @given(dates=pairs(dates(), sorted=True), func=sampled_from([repr, str]))
    def test_repr_date(
        self, *, dates: tuple[dt.date, dt.date], func: Callable[..., str]
    ) -> None:
        start, end = dates
        period = Period(start, end)
        result = func(period)
        assert search(r"^Period\(\d{4}-\d{2}-\d{2}, \d{4}-\d{2}-\d{2}\)$", result)

    @given(data=data(), time_zone=timezones(), func=sampled_from([repr, str]))
    def test_repr_datetime_same_time_zone(
        self, *, data: DataObject, time_zone: ZoneInfo, func: Callable[..., str]
    ) -> None:
        datetimes = data.draw(pairs(zoned_datetimes(time_zone=time_zone), sorted=True))
        start, end = datetimes
        period = Period(start, end)
        result = func(period)
        assert search(
            r"^Period\(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,6})?, \d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,6})?, .+\)$",
            result,
        )

    @given(
        datetimes=pairs(zoned_datetimes(time_zone=timezones()), sorted=True),
        time_zones=sets(timezones(), min_size=2, max_size=2),
        func=sampled_from([repr, str]),
    )
    def test_repr_datetime_different_time_zone(
        self,
        *,
        datetimes: tuple[dt.datetime, dt.datetime],
        time_zones: set[ZoneInfo],
        func: Callable[..., str],
    ) -> None:
        start, end = datetimes
        time_zone1, time_zone2 = time_zones
        period = Period(start.astimezone(time_zone1), end.astimezone(time_zone2))
        result = func(period)
        assert search(
            r"^Period\(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,6})?[\+-]\d{2}:\d{2}(:\d{2})?\[.+\], \d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,6})?[\+-]\d{2}:\d{2}(:\d{2})?\[.+\]\)$",
            result,
        )

    @given(dates=pairs(dates(), sorted=True), extra=booleans())
    def test_repr_sub_classes(
        self, *, dates: tuple[dt.date, dt.date], extra: bool
    ) -> None:
        start, end = dates

        @dataclass
        class SubPeriod(Period[_TPeriod]):
            extra: bool

        period = SubPeriod(start, end, extra)
        result = repr(period)
        assert search(
            r"SubPeriod\(start=datetime\.date\(\d{1,4}, \d{1,2}, \d{1,2}\), end=datetime\.date\(\d{1,4}, \d{1,2}, \d{1,2}\), extra=(?:True|False)\)$",
            result,
        )

    @given(dates1=pairs(dates(), sorted=True), dates2=pairs(dates(), sorted=True))
    def test_sortable(
        self, *, dates1: tuple[dt.date, dt.date], dates2: tuple[dt.date, dt.date]
    ) -> None:
        start1, end1 = dates1
        start2, end2 = dates2
        period1 = Period(start1, end1)
        period2 = Period(start2, end2)
        _ = sorted([period1, period2])

    @given(dates=pairs(dates(), sorted=True), duration=timedeltas())
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    def test_sub(
        self, *, dates: tuple[dt.date, dt.date], duration: dt.timedelta
    ) -> None:
        start, end = dates
        with assume_does_not_raise(OverflowError):
            adj_start, adj_end = start - duration, end - duration
        period = Period(start, end)
        result = period - duration
        expected = Period(adj_start, adj_end)
        assert result == expected

    @given(data=data(), time_zone=timezones())
    def test_time_zone(self, *, data: DataObject, time_zone: ZoneInfo) -> None:
        datetimes = data.draw(pairs(zoned_datetimes(time_zone=time_zone), sorted=True))
        start, end = datetimes
        period = Period(start, end)
        assert period.time_zone is time_zone

    @given(dates=pairs(dates(), sorted=True))
    def test_to_dict(self, *, dates: tuple[dt.date, dt.date]) -> None:
        start, end = dates
        period = Period(start, end)
        result = period.to_dict()
        expected = {"start": start, "end": end}
        assert result == expected

    @given(dates=pairs(dates(), sorted=True), time_zone=timezones())
    def test_error_as_time_zone_inapplicable(
        self, *, dates: tuple[dt.date, dt.date], time_zone: ZoneInfo
    ) -> None:
        start, end = dates
        period = Period(start, end)
        with raises(
            _PeriodAsTimeZoneInapplicableError,
            match="Period of dates does not have a timezone attribute",
        ):
            _ = period.astimezone(time_zone)

    @given(data=data(), date=dates(), datetime=zoned_datetimes(time_zone=timezones()))
    def test_error_date_and_datetime_mix(
        self, *, data: DataObject, date: dt.date, datetime: dt.datetime
    ) -> None:
        start, end = data.draw(permutations([date, datetime]))
        with raises(
            _PeriodDateAndDateTimeMixedError,
            match=r"Invalid period; got date and datetime mix \(.*, .*\)",
        ):
            _ = Period(start, end)

    @given(datetimes=pairs(datetimes() | zoned_datetimes(time_zone=timezones())))
    def test_error_naive_datetime(
        self, *, datetimes: tuple[dt.datetime, dt.datetime]
    ) -> None:
        start, end = datetimes
        _ = assume((start.tzinfo is None) or (end.tzinfo is None))
        with raises(
            _PeriodNaiveDateTimeError,
            match=r"Invalid period; got naive datetime\(s\) \(.*, .*\)",
        ):
            _ = Period(start, end)

    @given(dates=pairs(dates(), unique=True, sorted=True))
    def test_error_invalid_dates(self, *, dates: tuple[dt.date, dt.date]) -> None:
        start, end = dates
        with raises(_PeriodInvalidError, match="Invalid period; got .* > .*"):
            _ = Period(end, start)

    @given(
        datetimes=pairs(
            zoned_datetimes(time_zone=timezones()), unique=True, sorted=True
        )
    )
    def test_error_invalid_datetimes(
        self, *, datetimes: tuple[dt.datetime, dt.datetime]
    ) -> None:
        start, end = datetimes
        with raises(_PeriodInvalidError, match="Invalid period; got .* > .*"):
            _ = Period(end, start)

    @given(dates=pairs(dates(), sorted=True), duration=timedeltas(min_value=ZERO_TIME))
    def test_error_req_duration(
        self, *, dates: tuple[dt.date, dt.date], duration: dt.timedelta
    ) -> None:
        start, end = dates
        _ = assume(end - start != duration)
        with raises(
            _PeriodReqDurationError, match="Period must have duration .*; got .*"
        ):
            _ = Period(start, end, req_duration=duration)

    @given(
        dates=pairs(dates(), sorted=True), min_duration=timedeltas(min_value=ZERO_TIME)
    )
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    def test_error_min_duration(
        self, *, dates: tuple[dt.date, dt.date], min_duration: dt.timedelta
    ) -> None:
        start, end = dates
        _ = assume(end - start < min_duration)
        with raises(
            _PeriodMinDurationError, match="Period must have min duration .*; got .*"
        ):
            _ = Period(start, end, min_duration=min_duration)

    @given(
        dates=pairs(dates(), sorted=True), max_duration=timedeltas(max_value=ZERO_TIME)
    )
    def test_error_max_duration(
        self, *, dates: tuple[dt.date, dt.date], max_duration: dt.timedelta
    ) -> None:
        start, end = dates
        _ = assume(end - start > max_duration)
        with raises(
            _PeriodMaxDurationError,
            match="Period must have duration at most .*; got .*",
        ):
            _ = Period(start, end, max_duration=max_duration)

    @given(
        datetime=zoned_datetimes(time_zone=timezones()),
        dates=pairs(dates(), sorted=True),
    )
    def test_error_date_contains_datetime(
        self, *, datetime: dt.datetime, dates: tuple[dt.date, dt.date]
    ) -> None:
        start, end = dates
        period = Period(start, end)
        with raises(
            _PeriodDateContainsDateTimeError,
            match="Period of dates cannot contain datetimes",
        ):
            _ = datetime in period

    @given(
        date=dates(),
        datetimes=pairs(zoned_datetimes(time_zone=timezones()), sorted=True),
    )
    def test_error_datetime_contains_date(
        self, *, date: dt.datetime, datetimes: tuple[dt.date, dt.date]
    ) -> None:
        start, end = datetimes
        period = Period(start, end)
        with raises(
            _PeriodDateTimeContainsDateError,
            match="Period of datetimes cannot contain dates",
        ):
            _ = date in period

    @given(dates=pairs(dates(), sorted=True))
    def test_error_time_zone_inapplicable(
        self, *, dates: tuple[dt.date, dt.date]
    ) -> None:
        start, end = dates
        period = Period(start, end)
        with raises(
            _PeriodTimeZoneInapplicableError,
            match="Period of dates does not have a timezone attribute",
        ):
            _ = period.time_zone

    @given(
        datetimes=pairs(zoned_datetimes(time_zone=timezones()), sorted=True),
        time_zones=sets(timezones(), min_size=2, max_size=2),
    )
    def test_error_time_zone_non_unique(
        self, *, datetimes: tuple[dt.datetime, dt.datetime], time_zones: set[ZoneInfo]
    ) -> None:
        start, end = datetimes
        time_zone1, time_zone2 = time_zones
        period = Period(start.astimezone(time_zone1), end.astimezone(time_zone2))
        with raises(
            _PeriodTimeZoneNonUniqueError,
            match="Period must contain exactly one time zone; got .* and .*",
        ):
            _ = period.time_zone



from __future__ import annotations

from asyncio import TaskGroup, run
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from inspect import iscoroutinefunction
from sys import version_info
from typing import TYPE_CHECKING, Any, cast, override

from utilities.logging import get_logger
from utilities.types import Coroutine1, LoggerOrName

if TYPE_CHECKING:
    from collections.abc import Iterable
    from types import TracebackType

    from utilities.types import MaybeCoroutine1, StrMapping


VERSION_MAJOR_MINOR = (version_info.major, version_info.minor)


def make_except_hook(
    *,
    logger: LoggerOrName | None = None,
    message: object = "",
    extra: StrMapping | None = None,
    callbacks: Iterable[Callable[[], MaybeCoroutine1[None]]] | None = None,
) -> Callable[
    [type[BaseException] | None, BaseException | None, TracebackType | None], None
]:
    """Create an exception hook with various features."""
    return partial(
        _make_except_hook_inner,
        logger=logger,
        message=message,
        extra=extra,
        callbacks=callbacks,
    )


def _make_except_hook_inner(
    exc_type: type[BaseException] | None,
    exc_val: BaseException | None,
    traceback: TracebackType | None,
    /,
    *,
    logger: LoggerOrName | None = None,
    message: object = "",
    extra: StrMapping | None = None,
    callbacks: Iterable[Callable[[], MaybeCoroutine1[None]]] | None = None,
) -> None:
    """Exception hook to log the traceback."""
    _ = (exc_type, traceback)
    if exc_val is None:
        raise MakeExceptHookError
    logger_use = get_logger(logger=logger)
    exc_info = (exc_type, exc_val, traceback)
    logger_use.exception(message, exc_info=cast(Any, exc_info), extra=extra)
    async_callbacks: list[Callable[[], Coroutine1[None]]] = []
    if callbacks is not None:
        for callback in callbacks:
            if not iscoroutinefunction(callback):
                cast(Callable[[], None], callback)()
            else:  # skipif-ci
                async_callback = cast(Callable[[], Coroutine1[None]], callback)
                async_callbacks.append(async_callback)
    if len(async_callbacks) >= 1:  # skipif-ci
        run(_run_async_callbacks(async_callbacks))


@dataclass(kw_only=True, slots=True)
class MakeExceptHookError(Exception):
    @override
    def __str__(self) -> str:
        return "No exception to log"


async def _run_async_callbacks(
    callbacks: Iterable[Callable[[], Coroutine1[None]]], /
) -> None:
    """Run all asynchronous callbacks."""
    async with TaskGroup() as tg:  # skipif-ci
        for callback in callbacks:
            _ = tg.create_task(callback())


__all__ = ["VERSION_MAJOR_MINOR", "MakeExceptHookError", "make_except_hook"]



from __future__ import annotations

from asyncio import sleep
from itertools import chain
from os import getpid
from time import time_ns
from typing import TYPE_CHECKING, Any, Literal, cast, overload, override

from hypothesis import Phase, assume, given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    floats,
    integers,
    lists,
    none,
    sampled_from,
    sets,
    tuples,
    uuids,
)
from pytest import mark, param, raises
from sqlalchemy import Boolean, Column, Integer, MetaData, Table, select
from sqlalchemy.exc import DatabaseError, OperationalError, ProgrammingError
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column

from utilities.hypothesis import (
    int32s,
    pairs,
    settings_with_reduced_examples,
    sqlalchemy_engines,
    temp_paths,
)
from utilities.iterables import one
from utilities.modules import is_installed
from utilities.sqlalchemy import (
    CheckEngineError,
    Dialect,
    GetTableError,
    InsertItemsError,
    TablenameMixin,
    TableOrORMInstOrClass,
    Upserter,
    UpsertItemsError,
    _get_dialect,
    _get_dialect_max_params,
    _InsertItem,
    _is_pair_of_sequence_of_tuple_or_string_mapping_and_table,
    _is_pair_of_str_mapping_and_table,
    _is_pair_of_tuple_and_table,
    _is_pair_of_tuple_or_str_mapping_and_table,
    _map_mapping_to_table,
    _MapMappingToTableExtraColumnsError,
    _MapMappingToTableSnakeMapEmptyError,
    _MapMappingToTableSnakeMapNonUniqueError,
    _normalize_insert_item,
    _normalize_upsert_item,
    _NormalizedItem,
    _NormalizeInsertItemError,
    _orm_inst_to_dict,
    _prepare_insert_or_upsert_items,
    _prepare_insert_or_upsert_items_merge_items,
    _PrepareInsertOrUpsertItemsError,
    _SelectedOrAll,
    _tuple_to_mapping,
    check_engine,
    columnwise_max,
    columnwise_min,
    create_async_engine,
    ensure_tables_created,
    ensure_tables_dropped,
    get_chunk_size,
    get_column_names,
    get_columns,
    get_primary_key_values,
    get_table,
    get_table_name,
    hash_primary_key_values,
    insert_items,
    is_orm,
    is_table_or_orm,
    migrate_data,
    selectable_to_string,
    upsert_items,
    yield_primary_key_columns,
)
from utilities.text import strip_and_dedent
from utilities.typing import get_args

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path

    from utilities.types import StrMapping


def _table_names() -> SearchStrategy[str]:
    now = time_ns()
    return uuids().map(lambda x: f"test_{now}_{getpid()}_{str(x).replace('-', '')}")


@overload
def _upsert_triples(
    *, nullable: Literal[True]
) -> SearchStrategy[tuple[int, bool, bool]]: ...
@overload
def _upsert_triples(
    *, nullable: bool = ...
) -> SearchStrategy[tuple[int, bool, bool | None]]: ...
def _upsert_triples(
    *, nullable: bool = False
) -> SearchStrategy[tuple[int, bool, bool | None]]:
    elements = booleans()
    if nullable:
        elements |= none()
    return tuples(int32s(), booleans(), elements)


def _upsert_lists(
    *, nullable: bool = False, min_size: int = 0, max_size: int | None = None
) -> SearchStrategy[list[tuple[int, bool, bool | None]]]:
    return lists(
        _upsert_triples(nullable=nullable),
        min_size=min_size,
        max_size=max_size,
        unique_by=lambda x: x[0],
    )


class TestCheckEngine:
    @given(data=data())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_main(self, *, data: DataObject) -> None:
        engine = await sqlalchemy_engines(data)
        await check_engine(engine)

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_num_tables_pass(self, *, data: DataObject, name: str) -> None:
        table = Table(name, MetaData(), Column("id", Integer, primary_key=True))
        engine = await sqlalchemy_engines(data, table)
        await ensure_tables_created(engine, table)
        match _get_dialect(engine):
            case "sqlite":
                expected = 1
            case "postgresql":
                expected = (int(1e6), 1.0)
            case _ as dialect:
                raise NotImplementedError(dialect)
        await check_engine(engine, num_tables=expected)

    @given(data=data())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_num_tables_error(self, *, data: DataObject) -> None:
        engine = await sqlalchemy_engines(data)
        with raises(CheckEngineError, match=r".* must have 100000 table\(s\); got .*"):
            await check_engine(engine, num_tables=100000)


class TestColumnwiseMinMax:
    @given(
        data=data(),
        name=_table_names(),
        values=sets(pairs(integers(0, 10) | none()), min_size=1),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_main(
        self, *, data: DataObject, name: str, values: set[tuple[int | None, int | None]]
    ) -> None:
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        engine = await sqlalchemy_engines(data, table)
        await insert_items(engine, ([{"x": x, "y": y} for x, y in values], table))
        sel = select(
            table.c["x"],
            table.c["y"],
            columnwise_min(table.c["x"], table.c["y"]).label("min_xy"),
            columnwise_max(table.c["x"], table.c["y"]).label("max_xy"),
        )
        async with engine.begin() as conn:
            res = (await conn.execute(sel)).all()
        assert len(res) == len(values)
        for x, y, min_xy, max_xy in res:
            if (x is None) and (y is None):
                assert min_xy is None
                assert max_xy is None
            elif (x is not None) and (y is None):
                assert min_xy == x
                assert max_xy == x
            elif (x is None) and (y is not None):
                assert min_xy == y
                assert max_xy == y
            else:
                assert min_xy == min(x, y)
                assert max_xy == max(x, y)

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_label(self, *, data: DataObject, name: str) -> None:
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
        )
        engine = await sqlalchemy_engines(data, table)
        await ensure_tables_created(engine, table)
        sel = select(columnwise_min(table.c.x, table.c.x))
        async with engine.begin() as conn:
            _ = (await conn.execute(sel)).all()


class TestCreateAsyncEngine:
    @given(temp_path=temp_paths())
    @settings_with_reduced_examples(phases={Phase.generate})
    def test_async(self, *, temp_path: Path) -> None:
        engine = create_async_engine("sqlite+aiosqlite", database=temp_path.name)
        assert isinstance(engine, AsyncEngine)

    @given(temp_path=temp_paths())
    @settings_with_reduced_examples(phases={Phase.generate})
    def test_query(self, *, temp_path: Path) -> None:
        engine = create_async_engine(
            "sqlite+aiosqlite",
            database=temp_path.name,
            query={"arg1": "value1", "arg2": ["value2"]},
        )
        assert isinstance(engine, AsyncEngine)


class TestEnsureTablesCreated:
    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_table(self, *, data: DataObject, name: str) -> None:
        table = Table(name, MetaData(), Column("id_", Integer, primary_key=True))
        engine = await sqlalchemy_engines(data, table)
        await self._run_test(engine, table)

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_mapped_class(self, *, data: DataObject, name: str) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = name

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        engine = await sqlalchemy_engines(data, Example)
        await self._run_test(engine, Example)

    async def _run_test(
        self, engine: AsyncEngine, table_or_orm: TableOrORMInstOrClass, /
    ) -> None:
        for _ in range(2):
            await ensure_tables_created(engine, table_or_orm)
        sel = select(get_table(table_or_orm))
        async with engine.begin() as conn:
            _ = (await conn.execute(sel)).all()


class TestEnsureTablesDropped:
    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_table(self, *, data: DataObject, name: str) -> None:
        table = Table(name, MetaData(), Column("id_", Integer, primary_key=True))
        engine = await sqlalchemy_engines(data, table)
        await self._run_test(engine, table)

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_mapped_class(self, *, data: DataObject, name: str) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = name

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        engine = await sqlalchemy_engines(data, Example)
        await self._run_test(engine, Example)

    async def _run_test(
        self, engine: AsyncEngine, table_or_orm: TableOrORMInstOrClass, /
    ) -> None:
        for _ in range(2):
            await ensure_tables_dropped(engine, table_or_orm)
        sel = select(get_table(table_or_orm))
        with raises(DatabaseError):
            async with engine.begin() as conn:
                _ = await conn.execute(sel)


class TestGetChunkSize:
    @given(data=data(), chunk_size_frac=floats(0.0, 1.0), scaling=floats(0.1, 10.0))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_main(
        self, *, data: DataObject, chunk_size_frac: float, scaling: float
    ) -> None:
        engine = await sqlalchemy_engines(data)
        result = get_chunk_size(
            engine, chunk_size_frac=chunk_size_frac, scaling=scaling
        )
        assert result >= 1


class TestGetColumnNames:
    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        self._run_test(table)

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_orm: TableOrORMInstOrClass, /) -> None:
        assert get_column_names(table_or_orm) == ["id_"]


class TestGetColumns:
    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        self._run_test(table)

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_orm: TableOrORMInstOrClass, /) -> None:
        columns = get_columns(table_or_orm)
        assert isinstance(columns, list)
        assert len(columns) == 1
        assert isinstance(columns[0], Column)


class TestGetDialect:
    @mark.skipif(condition=not is_installed("pyodbc"), reason="'pyodbc' not installed")
    def test_mssql(self) -> None:
        engine = create_async_engine("mssql")
        assert _get_dialect(engine) == "mssql"

    @mark.skipif(
        condition=not is_installed("mysqldb"), reason="'mysqldb' not installed"
    )
    def test_mysql(self) -> None:
        engine = create_async_engine("mysql")
        assert _get_dialect(engine) == "mysql"

    @mark.skipif(
        condition=not is_installed("oracledb"), reason="'oracledb' not installed"
    )
    def test_oracle(self) -> None:
        engine = create_async_engine("oracle+oracledb")
        assert _get_dialect(engine) == "oracle"

    @mark.skipif(
        condition=not is_installed("asyncpg"), reason="'asyncpg' not installed"
    )
    def test_postgres(self) -> None:
        engine = create_async_engine("postgresql+asyncpg")
        assert _get_dialect(engine) == "postgresql"

    @mark.skipif(
        condition=not is_installed("aiosqlite"), reason="'asyncpg' not installed"
    )
    def test_sqlite(self) -> None:
        engine = create_async_engine("sqlite+aiosqlite")
        assert _get_dialect(engine) == "sqlite"


class TestGetDialectMaxParams:
    @given(dialect=sampled_from(get_args(Dialect)))
    def test_max_params(self, *, dialect: Dialect) -> None:
        max_params = _get_dialect_max_params(dialect)
        assert isinstance(max_params, int)


class TestGetPrimaryKeyValues:
    @given(id1=integers(), id2=integers(), value=booleans())
    def test_main(self, *, id1: int, id2: int, value: bool) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id1: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            id2: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            value: Mapped[bool] = mapped_column(Boolean, kw_only=True, nullable=False)

        obj = Example(id1=id1, id2=id2, value=value)
        result = get_primary_key_values(obj)
        expected = (id1, id2)
        assert result == expected


class TestGetTable:
    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        result = get_table(table)
        assert result is table

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        result = get_table(Example)
        expected = Example.__table__
        assert result is expected

    def test_instance_of_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        obj = Example(id=1)
        result = get_table(obj)
        expected = Example.__table__
        assert result is expected

    def test_error(self) -> None:
        with raises(
            GetTableError, match="Object .* must be a Table or mapped class; got .*"
        ):
            _ = get_table(cast(Any, type(None)))


class TestGetTableName:
    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        result = get_table_name(table)
        expected = "example"
        assert result == expected

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        result = get_table_name(Example)
        expected = "example"
        assert result == expected


class TestHashPrimaryKeyValues:
    @given(id1=integers(), id2=integers(), value=booleans())
    def test_main(self, *, id1: int, id2: int, value: bool) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id1: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            id2: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            value: Mapped[bool] = mapped_column(Boolean, kw_only=True, nullable=False)

            @override
            def __hash__(self) -> int:
                return hash_primary_key_values(self)

        obj = Example(id1=id1, id2=id2, value=value)
        result = hash(obj)
        expected = hash((id1, id2))
        assert result == expected


class TestInsertItems:
    @given(
        data=data(),
        name=_table_names(),
        case=sampled_from(["tuple", "dict"]),
        id_=integers(0, 10),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_pair_of_obj_and_table(
        self, *, data: DataObject, name: str, case: Literal["tuple", "dict"], id_: int
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        match case:
            case "tuple":
                item = (id_,), table
            case "dict":
                item = {"id_": id_}, table
        await self._run_test(engine, table, {id_}, item)

    @given(
        data=data(),
        name=_table_names(),
        case=sampled_from([
            "pair-list-of-tuples",
            "pair-list-of-dicts",
            "list-of-pair-of-tuples",
            "list-of-pair-of-dicts",
        ]),
        ids=sets(integers(0, 10), min_size=1),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_pair_of_objs_and_table_or_list_of_pairs_of_objs_and_table(
        self,
        *,
        data: DataObject,
        name: str,
        case: Literal[
            "pair-list-of-tuples",
            "pair-list-of-dicts",
            "list-of-pair-of-tuples",
            "list-of-pair-of-dicts",
        ],
        ids: set[int],
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        match case:
            case "pair-list-of-tuples":
                item = [((id_,)) for id_ in ids], table
            case "pair-list-of-dicts":
                item = [({"id_": id_}) for id_ in ids], table
            case "list-of-pair-of-tuples":
                item = [((id_,), table) for id_ in ids]
            case "list-of-pair-of-dicts":
                item = [({"id_": id_}, table) for id_ in ids]
        await self._run_test(engine, table, ids, item)

    @given(
        data=data(),
        name=_table_names(),
        ids=sets(integers(0, 1000), min_size=10, max_size=100),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_many_items(
        self, *, data: DataObject, name: str, ids: set[int]
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        await self._run_test(engine, table, ids, [({"id_": id_}, table) for id_ in ids])

    @given(data=data(), name=_table_names(), id_=integers(0, 10))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_mapped_class(self, *, data: DataObject, name: str, id_: int) -> None:
        cls = self._make_mapped_class(name)
        engine = await sqlalchemy_engines(data, cls)
        await self._run_test(engine, cls, {id_}, cls(id_=id_))

    @given(data=data(), name=_table_names(), ids=sets(integers(0, 10), min_size=1))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_mapped_classes(
        self, *, data: DataObject, name: str, ids: set[int]
    ) -> None:
        cls = self._make_mapped_class(name)
        engine = await sqlalchemy_engines(data, cls)
        await self._run_test(engine, cls, ids, [cls(id_=id_) for id_ in ids])

    @given(data=data(), name=_table_names(), id_=integers(0, 10))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_snake(self, *, data: DataObject, name: str, id_: int) -> None:
        table = self._make_table(name, title=True)
        engine = await sqlalchemy_engines(data, table)
        item = {data.draw(sampled_from(["Id_", "id_"])): id_}, table
        await self._run_test(engine, table, {id_}, item, snake=True)

    @given(data=data(), name=_table_names(), id_=integers(0, 10))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_assume_table_exists(
        self, *, data: DataObject, name: str, id_: int
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        with raises(
            (OperationalError, ProgrammingError), match="(no such table|does not exist)"
        ):
            await insert_items(engine, ({"id_": id_}, table), assume_tables_exist=True)

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_error(self, *, data: DataObject, name: str) -> None:
        cls = self._make_mapped_class(name)
        engine = await sqlalchemy_engines(data, cls)
        with raises(InsertItemsError, match="Item must be valid; got None"):
            await self._run_test(engine, cls, set(), cast(Any, None))

    def _make_table(self, name: str, /, *, title: bool = False) -> Table:
        return Table(
            name,
            MetaData(),
            Column("Id_" if title else "id_", Integer, primary_key=True),
        )

    def _make_mapped_class(self, name: str, /) -> type[DeclarativeBase]:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = name

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        return Example

    async def _run_test(
        self,
        engine: AsyncEngine,
        table_or_orm: TableOrORMInstOrClass,
        ids: set[int],
        /,
        *items: _InsertItem,
        snake: bool = False,
    ) -> None:
        await insert_items(engine, *items, snake=snake)
        sel = select(get_table(table_or_orm).c["Id_" if snake else "id_"])
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).scalars().all()
        assert set(results) == ids


class TestIsPairOfSequenceOfTupleOrStringMappingAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(([(1, 2, 3)], Table("example", MetaData())), True),
            param(([{"a": 1, "b": 2, "c": 3}], Table("example", MetaData())), True),
            param(
                ([(1, 2, 3), {"a": 1, "b": 2, "c": 3}], Table("example", MetaData())),
                True,
            ),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_sequence_of_tuple_or_string_mapping_and_table(obj)
        assert result is expected


class TestIsPairOfStrMappingAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(((1, 2, 3), Table("example", MetaData())), False),
            param(({"a": 1, "b": 2, "c": 3}, Table("example", MetaData())), True),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_str_mapping_and_table(obj)
        assert result is expected


class TestIsPairOfTupleAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(((1, 2, 3), Table("example", MetaData())), True),
            param(({"a": 1, "b": 2, "c": 3}, Table("example", MetaData())), False),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_tuple_and_table(obj)
        assert result is expected


class TestIsPairOfTupleStrMappingAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(((1, 2, 3), Table("example", MetaData())), True),
            param(({"a": 1, "b": 2, "c": 3}, Table("example", MetaData())), True),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_tuple_or_str_mapping_and_table(obj)
        assert result is expected


class TestIsORM:
    def test_orm_inst(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        obj = Example(id_=1)
        assert is_table_or_orm(obj)

    def test_orm_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        assert is_table_or_orm(Example)

    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        assert not is_orm(table)

    def test_none(self) -> None:
        assert not is_orm(None)


class TestIsTableOrORM:
    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        assert is_table_or_orm(table)

    def test_orm_inst(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        obj = Example(id_=1)
        assert is_table_or_orm(obj)

    def test_orm_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        assert is_table_or_orm(Example)

    def test_other(self) -> None:
        assert not is_table_or_orm(None)


class TestMapMappingToTable:
    @given(id_=integers(0, 10), value=booleans())
    def test_main(self, *, id_: int, value: bool) -> None:
        mapping = {"id_": id_, "value": value}
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        result = _map_mapping_to_table(mapping, table)
        assert result == mapping

    @given(data=data(), id_=integers(0, 10), value=booleans())
    def test_snake(self, *, data: DataObject, id_: int, value: bool) -> None:
        mapping = {
            data.draw(sampled_from(["Id_", "id_"])): id_,
            data.draw(sampled_from(["Value", "value"])): value,
        }
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        result = _map_mapping_to_table(mapping, table, snake=True)
        expected = {"id_": id_, "value": value}
        assert result == expected

    @given(id_=integers(0, 10), value=booleans(), extra=booleans())
    def test_error_extra_columns(self, *, id_: int, value: bool, extra: bool) -> None:
        mapping = {"id_": id_, "value": value, "extra": extra}
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        with raises(
            _MapMappingToTableExtraColumnsError,
            match=r"Mapping .* must be a subset of table columns .*; got extra .*",
        ):
            _ = _map_mapping_to_table(mapping, table)

    @given(id_=integers(0, 10), value=booleans())
    def test_error_snake_empty_error(self, *, id_: int, value: bool) -> None:
        mapping = {"id_": id_, "invalid": value}
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        with raises(
            _MapMappingToTableSnakeMapEmptyError,
            match=r"Mapping .* must be a subset of table columns .*; cannot find column to map to 'invalid' modulo snake casing",
        ):
            _ = _map_mapping_to_table(mapping, table, snake=True)

    @given(id_=integers(0, 10), value=booleans())
    def test_error_snake_non_unique_error(self, *, id_: int, value: bool) -> None:
        mapping = {"id_": id_, "value": value}
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
            Column("Value", Boolean),
        )
        with raises(
            _MapMappingToTableSnakeMapNonUniqueError,
            match=r"Mapping .* must be a subset of table columns .*; found columns 'value', 'Value' and perhaps more to map to 'value' modulo snake casing",
        ):
            _ = _map_mapping_to_table(mapping, table, snake=True)


class TestMigrateData:
    @given(
        data=data(),
        names=sets(_table_names(), min_size=2, max_size=2),
        values=lists(
            tuples(integers(0, 10), booleans() | none()),
            min_size=1,
            unique_by=lambda x: x[0],
        ),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_main(
        self, *, data: DataObject, names: set[str], values: list[tuple[int, bool]]
    ) -> None:
        engine1 = await sqlalchemy_engines(data)
        name1, name2 = names
        table1 = self._make_table(name1)
        await insert_items(
            engine1, [({"id_": id_, "value": v}, table1) for id_, v in values]
        )
        async with engine1.begin() as conn:
            result1 = (await conn.execute(select(table1))).all()
        assert len(result1) == len(values)

        engine2 = await sqlalchemy_engines(data)
        table2 = self._make_table(name2)
        await migrate_data(table1, engine1, engine2, table_or_orm_to=table2)
        async with engine2.begin() as conn:
            result2 = (await conn.execute(select(table2))).all()
        assert len(result2) == len(values)

    def _make_table(self, name: str, /) -> Table:
        return Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )


class TestNormalizeInsertItem:
    @given(case=sampled_from(["tuple", "dict"]), id_=integers(0, 10))
    def test_pair_of_tuple_or_str_mapping_and_table(
        self, *, case: Literal["tuple", "dict"], id_: int
    ) -> None:
        table = self._table
        match case:
            case "tuple":
                item = (id_,), table
            case "dict":
                item = {"id_": id_}, table
        result = one(_normalize_insert_item(item))
        expected = _NormalizedItem(mapping={"id_": id_}, table=table)
        assert result == expected

    @given(case=sampled_from(["tuple", "dict"]), ids=sets(integers(0, 10)))
    def test_pair_of_list_of_tuples_or_str_mappings_and_table(
        self, *, case: Literal["tuple", "dict"], ids: set[int]
    ) -> None:
        table = self._table
        match case:
            case "tuple":
                item = [((id_,)) for id_ in ids], table
            case "dict":
                item = [({"id_": id_}) for id_ in ids], table
        result = list(_normalize_insert_item(item))
        expected = [_NormalizedItem(mapping={"id_": id_}, table=table) for id_ in ids]
        assert result == expected

    @given(case=sampled_from(["tuple", "dict"]), ids=sets(integers()))
    def test_list_of_pairs_of_objs_and_table(
        self, *, case: Literal["tuple", "dict"], ids: set[int]
    ) -> None:
        table = self._table
        match case:
            case "tuple":
                item = [(((id_,), table)) for id_ in ids]
            case "dict":
                item = [({"id_": id_}, table) for id_ in ids]
        result = list(_normalize_insert_item(item))
        expected = [_NormalizedItem(mapping={"id_": id_}, table=table) for id_ in ids]
        assert result == expected

    @given(id_=integers())
    def test_mapped_class(self, *, id_: int) -> None:
        cls = self._mapped_class
        result = one(_normalize_insert_item(cls(id_=id_)))
        expected = _NormalizedItem(mapping={"id_": id_}, table=get_table(cls))
        assert result == expected

    @given(ids=sets(integers(0, 10), min_size=1))
    def test_mapped_classes(self, *, ids: set[int]) -> None:
        cls = self._mapped_class
        result = list(_normalize_insert_item([cls(id_=id_) for id_ in ids]))
        expected = [
            _NormalizedItem(mapping={"id_": id_}, table=get_table(cls)) for id_ in ids
        ]
        assert result == expected

    @given(case=sampled_from(["tuple", "dict"]), id_=integers(0, 10))
    def test_snake(self, *, case: Literal["tuple", "dict"], id_: int) -> None:
        table = Table("example", MetaData(), Column("Id_", Integer, primary_key=True))
        match case:
            case "tuple":
                item = (id_,), table
            case "dict":
                item = {"id_": id_}, table
        result = one(_normalize_insert_item(item, snake=True))
        expected = _NormalizedItem(mapping={"Id_": id_}, table=table)
        assert result == expected

    @mark.parametrize(
        "item",
        [
            param((None,), id="tuple, not pair"),
            param(
                (None, Table("example", MetaData())), id="pair, first element invalid"
            ),
            param(((1, 2, 3), None), id="pair, second element invalid"),
            param([None], id="iterable, invalid"),
            param(None, id="outright invalid"),
        ],
    )
    def test_errors(self, *, item: Any) -> None:
        with raises(_NormalizeInsertItemError, match="Item must be valid; got .*"):
            _ = list(_normalize_insert_item(item))

    @property
    def _table(self) -> Table:
        return Table("example", MetaData(), Column("id_", Integer, primary_key=True))

    @property
    def _mapped_class(self) -> type[DeclarativeBase]:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        return Example


class TestORMInstToDict:
    @given(id_=integers())
    def test_main(self, *, id_: int) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        example = Example(id_=id_)
        result = _orm_inst_to_dict(example)
        expected = {"id_": id_}
        assert result == expected

    @given(id_=integers())
    def test_explicitly_named_column(self, *, id_: int) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = "example"

            ID: Mapped[int] = mapped_column(
                Integer, kw_only=True, primary_key=True, name="id"
            )

        example = Example(ID=id_)
        result = _orm_inst_to_dict(example)
        expected = {"id": id_}
        assert result == expected


class TestPrepareInsertOrUpsertItems:
    @given(
        data=data(),
        normalize_item=sampled_from([_normalize_insert_item, _normalize_upsert_item]),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_error(
        self, *, data: DataObject, normalize_item: Callable[[Any], Iterator[Any]]
    ) -> None:
        engine = await sqlalchemy_engines(data)
        with raises(
            _PrepareInsertOrUpsertItemsError, match="Item must be valid; got None"
        ):
            _ = _prepare_insert_or_upsert_items(
                normalize_item, engine, cast(Any, None), cast(Any, None)
            )


class TestPrepareInsertOrUpsertItemsMergeItems:
    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_main(self, *, data: DataObject, name: str) -> None:
        engine = await sqlalchemy_engines(data)
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )
        await ensure_tables_created(engine, table)
        items = [
            {"id_": 1, "value": True},
            {"id_": 1, "value": False},
            {"id_": 2, "value": False},
            {"id_": 2, "value": True},
        ]
        result = _prepare_insert_or_upsert_items_merge_items(table, items)
        expected = [{"id_": 1, "value": False}, {"id_": 2, "value": True}]
        assert result == expected
        async with engine.begin() as conn:
            _ = await conn.execute(table.insert().values(expected))

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_just_value(self, *, data: DataObject, name: str) -> None:
        engine = await sqlalchemy_engines(data)
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Integer),
        )
        await ensure_tables_created(engine, table)
        items = [{"value": 1}, {"value": 2}]
        result = _prepare_insert_or_upsert_items_merge_items(table, items)
        assert result == items
        async with engine.begin() as conn:
            _ = await conn.execute(table.insert().values(items))

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_autoincrement(self, *, data: DataObject, name: str) -> None:
        engine = await sqlalchemy_engines(data)
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("value", Integer),
        )
        await ensure_tables_created(engine, table)
        items = [{"value": 1}, {"value": 2}]
        result = _prepare_insert_or_upsert_items_merge_items(table, items)
        assert result == items
        async with engine.begin() as conn:
            _ = await conn.execute(table.insert().values(items))


class TestSelectableToString:
    @given(data=data())
    @settings_with_reduced_examples()
    async def test_main(self, *, data: DataObject) -> None:
        engine = await sqlalchemy_engines(data)
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )
        sel = select(table).where(table.c.value >= 1)
        result = selectable_to_string(sel, engine)
        expected = strip_and_dedent(
            """
                SELECT example.id_, example.value --
                FROM example --
                WHERE example.value >= 1
            """.replace("--\n", "\n")
        )
        assert result == expected


class TestTablenameMixin:
    def test_main(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass, TablenameMixin): ...

        class Example(Base):
            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        assert get_table_name(Example) == "example"


class TestTupleToMapping:
    @mark.parametrize(
        ("values", "expected"),
        [
            param((), {}),
            param((1,), {"id_": 1}),
            param((1, True), {"id_": 1, "value": True}),
            param((None, True), {"value": True}),
        ],
        ids=str,
    )
    def test_main(self, *, values: tuple[Any, ...], expected: StrMapping) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )
        result = _tuple_to_mapping(values, table)
        assert result == expected


class TestUpserter:
    @given(
        data=data(),
        name=_table_names(),
        triples=_upsert_lists(nullable=True, min_size=1),
    )
    @mark.flaky
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_main(
        self, *, data: DataObject, name: str, triples: list[tuple[int, bool, bool]]
    ) -> None:
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )
        engine = await sqlalchemy_engines(data, table)
        pairs = [(id_, init) for id_, init, _ in triples]
        async with Upserter(engine=engine) as upserter:
            upserter.enqueue((pairs, table))
            await sleep(0.2)
        sel = select(table)
        async with engine.begin() as conn:
            res = (await conn.execute(sel)).all()
        assert set(res) == set(pairs)


class TestUpsertItems:
    @given(data=data(), name=_table_names(), triple=_upsert_triples(nullable=True))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_pair_of_dict_and_table(
        self, *, data: DataObject, name: str, triple: tuple[int, bool, bool | None]
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        id_, init, post = triple
        init_item = {"id_": id_, "value": init}, table
        await self._run_test(engine, table, init_item, expected={(id_, init)})
        post_item = {"id_": id_, "value": post}, table
        _ = await self._run_test(
            engine, table, post_item, expected={(id_, init if post is None else post)}
        )

    @given(
        data=data(),
        name=_table_names(),
        triples=_upsert_lists(nullable=True, min_size=1),
        case=sampled_from(["pair-list-of-dicts", "list-of-pair-of-dicts"]),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_pair_of_list_of_dicts_and_table(
        self,
        *,
        data: DataObject,
        name: str,
        triples: list[tuple[int, bool, bool | None]],
        case: Literal["pair-list-of-dicts", "list-of-pair-of-dicts"],
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        match case:
            case "pair-list-of-dicts":
                init = (
                    [{"id_": id_, "value": init} for id_, init, _ in triples],
                    table,
                )
                post = (
                    [
                        {"id_": id_, "value": post}
                        for id_, _, post in triples
                        if post is not None
                    ],
                    table,
                )
            case "list-of-pair-of-dicts":
                init = [
                    ({"id_": id_, "value": init}, table) for id_, init, _ in triples
                ]
                post = [
                    ({"id_": id_, "value": post}, table)
                    for id_, _, post in triples
                    if post is not None
                ]
        init_expected = {(id_, init) for id_, init, _ in triples}
        _ = await self._run_test(engine, table, init, expected=init_expected)
        post_expected = {
            (id_, init if post is None else post) for id_, init, post in triples
        }
        _ = await self._run_test(engine, table, post, expected=post_expected)

    @given(data=data(), name=_table_names(), triple=_upsert_triples())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_mapped_class(
        self, *, data: DataObject, name: str, triple: tuple[int, bool, bool]
    ) -> None:
        cls = self._make_mapped_class(name)
        engine = await sqlalchemy_engines(data, cls)
        id_, init, post = triple
        _ = await self._run_test(
            engine, cls, cls(id_=id_, value=init), expected={(id_, init)}
        )
        _ = await self._run_test(
            engine, cls, cls(id_=id_, value=post), expected={(id_, post)}
        )

    @given(
        data=data(),
        name=_table_names(),
        triples=_upsert_lists(nullable=True, min_size=1),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_mapped_classes(
        self,
        *,
        data: DataObject,
        name: str,
        triples: list[tuple[int, bool, bool | None]],
    ) -> None:
        cls = self._make_mapped_class(name)
        engine = await sqlalchemy_engines(data, cls)
        init = [cls(id_=id_, value=init) for id_, init, _ in triples]
        init_expected = {(id_, init) for id_, init, _ in triples}
        _ = await self._run_test(engine, cls, init, expected=init_expected)
        post = [
            cls(id_=id_, value=post) for id_, _, post in triples if post is not None
        ]
        post_expected = {
            (id_, init if post is None else post) for id_, init, post in triples
        }
        _ = await self._run_test(engine, cls, post, expected=post_expected)

    @given(
        data=data(),
        name=_table_names(),
        id_=integers(0, 10),
        x_init=booleans(),
        x_post=booleans(),
        y=booleans(),
        selected_or_all=sampled_from(get_args(_SelectedOrAll)),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_sel_or_all(
        self,
        *,
        data: DataObject,
        name: str,
        selected_or_all: _SelectedOrAll,
        id_: int,
        x_init: bool,
        x_post: bool,
        y: bool,
    ) -> None:
        table = Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("x", Boolean, nullable=False),
            Column("y", Boolean, nullable=True),
        )
        engine = await sqlalchemy_engines(data, table)
        _ = await self._run_test(
            engine,
            table,
            ({"id_": id_, "x": x_init, "y": y}, table),
            selected_or_all=selected_or_all,
            expected={(id_, x_init, y)},
        )
        match selected_or_all:
            case "selected":
                expected = (id_, x_post, y)
            case "all":
                expected = (id_, x_post, None)
        _ = await self._run_test(
            engine,
            table,
            ({"id_": id_, "x": x_post}, table),
            selected_or_all=selected_or_all,
            expected={expected},
        )

    @given(data=data(), name=_table_names(), id_=integers(0, 10))
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_assume_table_exists(
        self, *, data: DataObject, name: str, id_: int
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        with raises((OperationalError, ProgrammingError)):
            await upsert_items(
                engine, ({"id_": id_, "value": True}, table), assume_tables_exist=True
            )

    @given(
        data=data(),
        name=_table_names(),
        id1=int32s(),
        id2=int32s(),
        value1=booleans() | none(),
        value2=booleans() | none(),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_both_nulls_and_non_nulls(
        self,
        *,
        data: DataObject,
        name: str,
        id1: int,
        id2: int,
        value1: bool | None,
        value2: bool | None,
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        _ = assume(id1 != id2)
        item = [{"id_": id1, "value": value1}, {"id_": id2, "value": value2}], table
        await upsert_items(engine, item)

    @given(
        data=data(),
        name=_table_names(),
        triples=_upsert_lists(nullable=True, min_size=1),
    )
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_multiple_elements_with_the_same_primary_key(
        self,
        *,
        data: DataObject,
        name: str,
        triples: list[tuple[int, bool, bool | None]],
    ) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        pairs = [
            ({"id_": id_, "value": init}, {"id_": id_, "value": post})
            for id_, init, post in triples
        ]
        item = list(chain.from_iterable(pairs)), table
        expected = {
            (id_, init if post is None else post) for id_, init, post in triples
        }
        await self._run_test(engine, table, item, expected=expected)

    @given(data=data(), name=_table_names())
    @settings_with_reduced_examples(phases={Phase.generate})
    async def test_error(self, *, data: DataObject, name: str) -> None:
        table = self._make_table(name)
        engine = await sqlalchemy_engines(data, table)
        with raises(UpsertItemsError, match="Item must be valid; got None"):
            _ = await self._run_test(engine, table, cast(Any, None))

    def _make_table(self, name: str, /) -> Table:
        return Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )

    def _make_mapped_class(self, name: str, /) -> type[DeclarativeBase]:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = name

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            value: Mapped[bool] = mapped_column(Boolean, kw_only=True, nullable=False)

        return Example

    async def _run_test(
        self,
        engine: AsyncEngine,
        table_or_orm: TableOrORMInstOrClass,
        /,
        *items: _InsertItem,
        selected_or_all: _SelectedOrAll = "selected",
        expected: set[tuple[Any, ...]] | None = None,
    ) -> None:
        await upsert_items(engine, *items, selected_or_all=selected_or_all)
        sel = select(get_table(table_or_orm))
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).all()
        if expected is not None:
            assert set(results) == expected


class TestYieldPrimaryKeyColumns:
    def test_main(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
            Column("id3", Integer),
        )
        result = list(yield_primary_key_columns(table))
        expected = [
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
        ]
        for c, e in zip(result, expected, strict=True):
            assert c.name == e.name
            assert c.primary_key == e.primary_key

    def test_autoincrement(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        result = list(yield_primary_key_columns(table, autoincrement=False))
        assert result == []



from __future__ import annotations

from enum import Enum, StrEnum, auto
from typing import TYPE_CHECKING, Literal

from hypothesis import given
from hypothesis.strategies import DataObject, data, sampled_from
from pytest import raises

from utilities.enum import (
    _EnsureEnumParseError,
    _EnsureEnumTypeEnumError,
    _ParseEnumByKindNonUniqueError,
    _ParseEnumGenericEnumEmptyError,
    _ParseEnumStrEnumEmptyError,
    _ParseEnumStrEnumNonUniqueError,
    ensure_enum,
    parse_enum,
)

if TYPE_CHECKING:
    from utilities.types import EnumOrStr


class TestParseEnum:
    @given(data=data())
    def test_generic_enum_case_insensitive(self, *, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        input_, expected = data.draw(
            sampled_from([
                ("true", Truth.true),
                ("TRUE", Truth.true),
                ("false", Truth.false),
                ("FALSE", Truth.false),
            ])
        )
        result = parse_enum(input_, Truth)
        assert result is expected

    @given(data=data())
    def test_generic_enum_case_sensitive(self, *, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        input_, expected = data.draw(
            sampled_from([("true", Truth.true), ("false", Truth.false)])
        )
        result = parse_enum(input_, Truth, case_sensitive=True)
        assert result is expected

    @given(data=data())
    def test_str_enum_case_insensitive(self, *, data: DataObject) -> None:
        class Truth(StrEnum):
            both = "both"
            true_by_name = "true_by_value"
            false_by_name = "false_by_value"

        input_, expected = data.draw(
            sampled_from([
                ("both", Truth.both),
                ("true_by_name", Truth.true_by_name),
                ("true_by_value", Truth.true_by_name),
                ("false_by_name", Truth.false_by_name),
                ("false_by_value", Truth.false_by_name),
            ])
        )
        input_ = data.draw(sampled_from([input_, input_.upper()]))
        result = parse_enum(input_, Truth)
        assert result is expected

    @given(data=data())
    def test_str_enum_case_sensitive(self, *, data: DataObject) -> None:
        class Truth(StrEnum):
            both = "both"
            true_by_name = "true_by_value"
            false_by_name = "false_by_value"

        input_, expected = data.draw(
            sampled_from([
                ("both", Truth.both),
                ("true_by_name", Truth.true_by_name),
                ("true_by_value", Truth.true_by_name),
                ("false_by_name", Truth.false_by_name),
                ("false_by_value", Truth.false_by_name),
            ])
        )
        result = parse_enum(input_, Truth, case_sensitive=True)
        assert result is expected

    def test_error_generic_enum_case_insensitive_empty(self) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        with raises(
            _ParseEnumGenericEnumEmptyError,
            match=r"^Enum 'Truth' member names do not contain 'invalid' \(modulo case\)",
        ):
            _ = parse_enum("invalid", Truth)

    @given(input_=sampled_from(["x", "X"]))
    def test_error_generic_enum_case_insensitive_non_unique(
        self, *, input_: Literal["x", "X"]
    ) -> None:
        class Example(Enum):
            x = auto()
            X = auto()

        with raises(
            _ParseEnumByKindNonUniqueError,
            match=r"^Enum 'Example' member names must contain '[xX]' exactly once \(modulo case\); got 'x', 'X' and perhaps more",
        ):
            _ = parse_enum(input_, Example)

    def test_error_generic_enum_case_sensitive_empty(self) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        with raises(
            _ParseEnumGenericEnumEmptyError,
            match=r"^Enum 'Truth' member names do not contain 'invalid'",
        ):
            _ = parse_enum("invalid", Truth, case_sensitive=True)

    def test_error_str_enum_case_insensitive_empty(self) -> None:
        class Truth(StrEnum):
            true = "TRUE"
            false = "FALSE"

        with raises(
            _ParseEnumStrEnumEmptyError,
            match=r"^StrEnum 'Truth' member names and values do not contain 'invalid' \(modulo case\)",
        ):
            _ = parse_enum("invalid", Truth)

    @given(data=data(), input_=sampled_from(["true", "false"]))
    def test_error_str_enum_case_insensitive_non_unique(
        self, *, data: DataObject, input_: Literal["true", "false"]
    ) -> None:
        class Truth(StrEnum):
            true = "FALSE"
            false = "TRUE"

        input_use = data.draw(sampled_from([input_, input_.upper()]))
        with raises(
            _ParseEnumStrEnumNonUniqueError,
            match=r"^StrEnum 'Truth' member names and values must contain '(true|false|TRUE|FALSE)' exactly once \(modulo case\); got '(true|false)' by name and '(true|false)' by value",
        ):
            _ = parse_enum(input_use, Truth)

    def test_error_str_enum_case_sensitive_empty(self) -> None:
        class Truth(StrEnum):
            true = "true"
            false = "false"

        with raises(
            _ParseEnumStrEnumEmptyError,
            match=r"^StrEnum 'Truth' member names and values do not contain 'invalid'",
        ):
            _ = parse_enum("invalid", Truth, case_sensitive=True)

    @given(input_=sampled_from(["true", "false"]))
    def test_error_str_enum_case_sensitive_non_unique(
        self, *, input_: Literal["true", "false"]
    ) -> None:
        class Truth(StrEnum):
            true = "false"
            false = "true"

        with raises(
            _ParseEnumStrEnumNonUniqueError,
            match=r"^StrEnum 'Truth' member names and values must contain '(true|false)' exactly once; got '(true|false)' by name and '(true|false)' by value",
        ):
            _ = parse_enum(input_, Truth, case_sensitive=True)

    @given(data=data())
    def test_ensure(self, *, data: DataObject) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        truth: Truth = data.draw(sampled_from(Truth))
        input_: EnumOrStr[Truth] = data.draw(sampled_from([truth, truth.name]))
        result = ensure_enum(input_, Truth)
        assert result is truth

    def test_ensure_none(self) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        result = ensure_enum(None, Truth)
        assert result is None

    @given(data=data())
    def test_error_ensure_type(self, *, data: DataObject) -> None:
        class Truth1(Enum):
            true1 = auto()
            false1 = auto()

        class Truth2(Enum):
            true2 = auto()
            false2 = auto()

        truth: Truth1 = data.draw(sampled_from(Truth1))
        with raises(_EnsureEnumTypeEnumError, match=".* is not an instance of .*"):
            _ = ensure_enum(truth, Truth2)

    def test_error_ensure_parse(self) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        with raises(
            _EnsureEnumParseError, match="Unable to ensure enum; got 'invalid'"
        ):
            _ = ensure_enum("invalid", Truth)



from __future__ import annotations

from asyncio import Task, create_task
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Self, override

from fastapi import FastAPI
from uvicorn import Config, Server

if TYPE_CHECKING:
    from types import TracebackType


_LOCALHOST = "localhost"


class _PingerReceiverApp(FastAPI):
    """App for the ping pinger."""

    @override
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)  # skipif-ci

        @self.get("/ping")  # skipif-ci
        def ping() -> str:
            return "pong"  # skipif-ci

        _ = ping  # skipif-ci


@dataclass(kw_only=True, slots=True)
class PingReceiver:
    """A ping receiver."""

    host: str = "localhost"
    port: int
    _app: _PingerReceiverApp = field(default_factory=_PingerReceiverApp)
    _config: Config = field(init=False)
    _server: Server = field(init=False)
    _task: Task[None] | None = None

    def __post_init__(self) -> None:
        self._config = Config(self._app, host=self.host, port=self.port)  # skipif-ci
        self._server = Server(self._config)  # skipif-ci

    @classmethod
    async def ping(cls, port: int, *, host: str = _LOCALHOST) -> bool:
        """Ping the receiver."""
        from httpx import AsyncClient, ConnectError  # skipif-ci

        url = f"http://{host}:{port}/ping"  # skipif-ci
        try:  # skipif-ci
            async with AsyncClient() as client:
                response = await client.get(url)
        except ConnectError:  # skipif-ci
            return False
        else:  # skipif-ci
            return response.status_code == 200

    async def __aenter__(self) -> Self:
        """Start the server."""
        self._task = create_task(self._server.serve())  # skipif-ci
        return self  # skipif-ci

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """Stop the server."""
        _ = (exc_type, exc_value, traceback)  # skipif-ci
        self._server.should_exit = True  # skipif-ci


__all__ = ["PingReceiver"]
